freeze;

//  Based on 
//  Cassels, J.W.S.: Rational quadratic forms 
//  London ; New York : Academic Press, 1978.


import "nrs.m": 
  NormResidueSymbolTableFor, 
  ConvertIndexToRC;


function MakeSolutionPrimitive4(v)

  //  transforms solution of the quadratic form such that 
  //  its components are coprime (no integer divides all of them) 

  if (v eq Vector([0,0,0,0])) then
    return v;
  end if;

  gcd := GCD(GCD(v[1], v[2]), GCD(v[3], v[4]));
  if (gcd ne 1) then
    v[1] := v[1] div gcd;
    v[2] := v[2] div gcd;
    v[3] := v[3] div gcd;
    v[4] := v[4] div gcd;
  end if;

  return v;
end function;


intrinsic IsIsotropic4(
  a1::RngIntElt, a2::RngIntElt, a3::RngIntElt, a4::RngIntElt :
  Al := "CremonaRusin") -> BoolElt, ModTupRngElt
{ Returns whether the quaternary quadratic form is isotropic over rationals 
 and isotropic vector with integral coordinates if so }

  // solve trivial case: zero occures on diagonal
  if (a1 eq 0) then
    return true, Vector([1,0,0,0]);
  elif (a2 eq 0) then
    return true, Vector([0,1,0,0]);
  elif (a3 eq 0) then
    return true, Vector([0,0,1,0]);
  elif (a4 eq 0) then
    return true, Vector([0,0,0,1]);
  end if;

  //  create set of all relevant primes:
  //  infinity, 2, every prime dividing a coefficient
  set_of_primes := {IntegerRing() | 0, 2};

  for i:=1 to 4 do
    if i eq 1 then a := a1; 
    elif i eq 2 then a := a2;
    elif i eq 3 then a := a3;
    else a := a4;
    end if;

    factor := Factorization(AbsoluteValue(a));
    len := #factor;
    for j:=1 to len do
      Include(~set_of_primes, factor[j][1]);
    end for;
  end for;

  //  solving...

  prime_list := [IntegerRing() |];
  nrs_a_list := [IntegerRing() |];
  nrs_b_list := [IntegerRing() |];
  a_rc_list := [];
  b_rc_list := [];

  p_power_prod := 1;
  record_num := 0;
  for prime in set_of_primes do
    if (prime ne 0) then
      record_num := record_num + 1;
    end if;

    a1_rc := ResidueClass(a1, prime);
    a2_rc := ResidueClass(a2, prime);
    b1_rc := ResidueClass(-a3, prime);
    b2_rc := ResidueClass(-a4, prime);

    nrs_a := NormResidueSymbol(a1_rc, a2_rc);
    nrs_b := NormResidueSymbol(b1_rc, b2_rc);

    minus1_rc := ResidueClass(-1, prime);

    a_rc := minus1_rc * (a1_rc * a2_rc);
    b_rc := minus1_rc * (b1_rc * b2_rc);

    if ((nrs_a ne nrs_b) and Equal(a_rc, b_rc)) then
      //  there is apparently no t_rc such that 
      //    NRS(t_rc, a_rc) = a_rc  and
      //    NRS(t_rc, b_rc) = b_rc
      return false, _;
    end if;

    if (prime ne 0) then
      prime_list[record_num] := prime;
      nrs_a_list[record_num] := nrs_a;
      nrs_b_list[record_num] := nrs_b;
      a_rc_list[record_num] := a_rc;
      b_rc_list[record_num] := b_rc;
    end if;

    a_rc_line := NormResidueSymbolTableFor(a_rc);
    b_rc_line := NormResidueSymbolTableFor(b_rc);

    length := OverDimension(a_rc_line);
    i := 0;
    found := false;
    while ((i lt length) and not found) do
      i := i+1;
      found := ((a_rc_line[i] eq nrs_a) and (b_rc_line[i] eq nrs_b));
    end while;

    t_rc := ConvertIndexToRC(i, prime);

    if (prime eq 0) then
      if (t_rc`nonsquare) then
         p_power_prod := -1 * p_power_prod;
      end if;
    elif (prime eq 2) then
      if (t_rc`gener_2) then
        p_power_prod := p_power_prod * 2;
      end if;
    else
      if (t_rc`withprime) then
        p_power_prod := p_power_prod * prime;
      end if;
    end if;
  end for;

  //  t should be of the form (-)p0 * product_of_some_primes_dividing_coeffs
  //  where p0 is a (special) prime.
  //  (-)product_of... is computed already
  //  suitable p0 is looked for by searching all primes
  found := false;
  p_zero := 0;
  while not found do
    p_zero := NextPrime(p_zero);
    t := p_zero * p_power_prod;
    for i := 1 to record_num do
      t_rc := ResidueClass(t, prime_list[i]);
      found := true;
      if (NormResidueSymbol(t_rc, a_rc_list[i]) ne nrs_a_list[i]) or
         (NormResidueSymbol(t_rc, b_rc_list[i]) ne nrs_b_list[i]) then
        found := false;
        break i;
      end if;
    end for;
  end while;

  //  solve ternary forms and combine the solutions
  //  these forms sould be always solvable
  ba, a_sol := IsIsotropic3(DiagonalMatrix(IntegerRing(), 3, [a1, a2, -t]): Al := Al);
  bb, b_sol := IsIsotropic3(DiagonalMatrix(IntegerRing(), 3, [a3, a4, t]): Al := Al);

  if ((a_sol[3] eq 0) and (b_sol[3] eq 0)) then
    solution := Vector([a_sol[1], a_sol[2], b_sol[1], b_sol[2]]);
  else 
    solution := Vector([a_sol[1]*b_sol[3], a_sol[2]*b_sol[3],
                        b_sol[1]*a_sol[3], b_sol[2]*a_sol[3]]);
  end if;

  return true, MakeSolutionPrimitive4(solution);
end intrinsic;


intrinsic IsIsotropic4(QF_M::AlgMatElt :
  Al := "CremonaRusin") -> BoolElt, ModTupRngElt
{ Returns whether the quaternary quadratic form is isotropic over rationals 
 and isotropic vector with integral coordinates if so }

  require IsSymmetric(QF_M): "Symmetric matrix required.";
  require NumberOfRows(QF_M) eq 4: "4 x 4 matrix required.";

  QF_M1, T, rank := OrthogonalizeGram(QF_M);

  b, v := IsIsotropic4(QF_M1[1][1], QF_M1[2][2], QF_M1[3][3], QF_M1[4][4] : Al := Al);
  if b then 
    return true, MakeSolutionPrimitive4(v * T);
  end if;

  return false, _;

end intrinsic;


