// not all fields of the record are always used:
// prime: used always; for infinity prime is set to 0
// if prime = infinity, used are
//   nonsquare
// if prime = 2, used are
//   gener_2, gener_minus1, gener_5
// if prime = any odd prime, used are
//   nonsquare, withprime
ResidueClassRec := recformat<
  prime: RngIntElt, 
  nonsquare, withprime: BoolElt, 
  gener_2, gener_minus1, gener_5: BoolElt>;

// **************************************************************
// ** Tables for Norm Residue Symbol:
// **   Cassels p.43-44
// **************************************************************
// for prime = 2
// the order of representatives: 1, 5, -1, -5, 2, 10, -2, -10
NRS_2 := Matrix(IntegerRing(), 8,8,
  [1, 1, 1, 1, 1, 1, 1, 1,
   1, 1, 1, 1,-1,-1,-1,-1,
   1, 1,-1,-1, 1, 1,-1,-1,
   1, 1,-1,-1,-1,-1, 1, 1,
   1,-1, 1,-1, 1,-1, 1,-1,
   1,-1, 1,-1,-1, 1,-1, 1,
   1,-1,-1, 1, 1,-1,-1, 1,
   1,-1,-1, 1,-1, 1, 1,-1]);

// for prime = infinity
// the order of representatives: 1, -1
NRS_infinity := Matrix(IntegerRing(), 2,2,
  [1, 1,
   1,-1]);

// for odd finite prime
// -1 is a square in p-adic numbers 
//   ... (iff p is congruent to 1 modulo 4
// the order of representatives: 1, r, prime, r*prime
// (r is not congruent to a square modulo prime)
NRS_odd_prime1 := Matrix(IntegerRing(), 4,4, 
  [1, 1, 1, 1,
   1, 1,-1,-1,
   1,-1, 1,-1,
   1,-1,-1, 1]);

// norm residue symbol for odd finite prime
// -1 is not square in p-adic numbers 
NRS_odd_prime2 := Matrix(IntegerRing(), 4,4, 
  [1, 1, 1, 1,
   1, 1,-1,-1,
   1,-1,-1, 1,
   1,-1, 1,-1]);


// **************************************************************
// ** Equal:
// **************************************************************
intrinsic Equal(a_rc::Rec, b_rc::Rec) -> BoolElt
{ compares two residue clases }

  if (a_rc`prime ne b_rc`prime) then
    return false;
  end if;

  if (a_rc`prime eq 0) then
    return (a_rc`nonsquare eq b_rc`nonsquare);
  elif (a_rc`prime eq 2) then
    return(
      (a_rc`gener_2 eq b_rc`gener_2) and
      (a_rc`gener_minus1 eq b_rc`gener_minus1) and
      (a_rc`gener_5 eq b_rc`gener_5));
  else
    return(
      (a_rc`nonsquare eq b_rc`nonsquare) and
      (a_rc`withprime eq b_rc`withprime));
  end if;
end intrinsic;


// **************************************************************
// ** '*':
// **************************************************************
intrinsic '*'(a_rc::Rec, b_rc::Rec) -> Rec
{ multiplies two residue clases }

  product := rec< ResidueClassRec | prime := a_rc`prime >;

  if (a_rc`prime eq 2) then
    product`gener_2 := a_rc`gener_2 xor b_rc`gener_2;
    product`gener_minus1 := a_rc`gener_minus1 xor b_rc`gener_minus1;
    product`gener_5 := a_rc`gener_5 xor b_rc`gener_5;
  elif (a_rc`prime eq 0) then
    product`nonsquare := a_rc`nonsquare xor b_rc`nonsquare;
  else
    product`nonsquare := a_rc`nonsquare xor b_rc`nonsquare; 
    product`withprime := a_rc`withprime xor b_rc`withprime;
  end if;

  return product;
end intrinsic;


// **************************************************************
// ** ConvertRCToIndex:
// **************************************************************
function ConvertRCToIndex(rc)
  if (rc`prime eq 0) then
    index := (rc`nonsquare) select 2 else 1;
  elif (rc`prime eq 2) then
    index := 1;
    if (rc`gener_2) then index := index + 4; end if;
    if (rc`gener_minus1) then index := index + 2; end if;
    if (rc`gener_5) then index := index + 1; end if;
  else
    index := 1;
    if (rc`nonsquare) then index := index + 1; end if;
    if (rc`withprime) then index := index + 2; end if;
  end if;

  return index;
end function;


// **************************************************************
// ** ConvertIndexToRC:
// **************************************************************
function ConvertIndexToRC(index, p)
  rc := rec<ResidueClassRec | prime := p>;

  if (p eq 0) then
    rc`nonsquare := index eq 2;
  elif (p eq 2) then
    rc`gener_2 := index gt 4;
    rc`gener_minus1 := (index mod 4) in {0,3};
    rc`gener_5 := (index mod 2) eq 0;
  else
    rc`nonsquare := (index mod 2) eq 0;
    rc`withprime := index gt 2;
  end if;

  return rc;
end function;


// **************************************************************
// ** Representative:
// **   "converse" function to ResidueClass
// **************************************************************
intrinsic Representative(rc::Rec) -> RngIntElt
{ integer representative of the residue class }
  if rc`prime eq 0 then
    rep := (rc`nonsquare) select -1 else 1;
  elif rc`prime eq 2 then
    rep := 1;
    if (rc`gener_2) then rep := 2*rep; end if;
    if (rc`gener_minus1) then rep := -1*rep; end if;
    if (rc`gener_5) then rep := 5*rep; end if;
  else
    rep := 1;
    if (rc`nonsquare) then
      FF := FiniteField(rc`prime);
      for i:=2 to rc`prime-1 do
        if not IsSquare(FF ! i) then
          rep := i * rep;
          break i;
        end if;
      end for;
    end if;
    if (rc`withprime) then rep := rc`prime * rep; end if;
  end if;

  return rep;
end intrinsic;


// **************************************************************
// ** NormResidueSymbol:
// **************************************************************
intrinsic NormResidueSymbol(a_rc::Rec, b_rc::Rec) -> RngIntElt
{ Hilbert Norm Residue Symbol }
  if (a_rc`prime ne b_rc`prime) then
    print "ERROR";
    return 0;
  end if;

  index_a := ConvertRCToIndex(a_rc);
  index_b := ConvertRCToIndex(b_rc);

  if (a_rc`prime eq 0) then
    return NRS_infinity[index_a][index_b];
  end if;

  if (a_rc`prime eq 2) then
    return NRS_2[index_a][index_b];
  end if;

  if (a_rc`prime mod 4 eq 1) then
    return NRS_odd_prime1[index_a, index_b];
  else 
    return NRS_odd_prime2[index_a, index_b];
  end if;
end intrinsic;


// **************************************************************
// ** NormResidueSymbolTableFor:
// **   returns the whole line from the table belonging to rc
// **************************************************************
function NormResidueSymbolTableFor(rc)
  index := ConvertRCToIndex(rc);

  if (rc`prime eq 0) then
    return NRS_infinity[index];
  end if;

  if (rc`prime eq 2) then
    return NRS_2[index];
  end if;

  if (rc`prime mod 4 eq 1) then
    return NRS_odd_prime1[index];
  else 
    return NRS_odd_prime2[index];
  end if;
end function;


// **************************************************************
// ** ResidueClass:
// **   "converse" function to Representative
// **   input: a - integer, p - prime or 0 (for infinity)
// **   output: clas of a in Q*p|((Q*p)^2)
// **************************************************************
intrinsic ResidueClass(a::RngIntElt, p::RngIntElt) -> Rec
{ residue class of the integer with the given prime }
  if (a eq 0) then
    print "non-zero input required";
    return 0;
  end if;

  if (p eq 0) then
    rc := rec<ResidueClassRec | 
      prime := 0, 
      nonsquare := (Sign(a) lt 0)>;
    return rc;
  end if;

  if ((not IsPrime(p)) or (p lt 0)) then
    print "positive prime required";
    return 0;
  end if;

  if (p eq 2) then
    rc := rec<ResidueClassRec | 
      prime := 2,
      gener_2 := false,
      gener_minus1 := false,
      gener_5 := false >;

    aux := a;
    while IsDivisibleBy(aux, 4) do
      aux := aux div 4;
    end while;
    if IsDivisibleBy(aux, 2) then
      rc`gener_2 := true;
      aux := aux div 2;
    end if;
    aux := aux mod 8;
    if (aux eq 3) then
      rc`gener_minus1 := true;
      rc`gener_5 := true;
    elif (aux eq 5) then
      rc`gener_5 := true;
    elif (aux eq 7) then
      rc`gener_minus1 := true;
    end if;

    return rc;
  end if;

  factor_a1, factor_a2 := SquarefreeFactorization(AbsoluteValue(a));
  new_a := Sign(a) * factor_a1;

  rc := rec<ResidueClassRec | prime := p, nonsquare := false, withprime := false>;
  if IsDivisibleBy(new_a, p) then
    rc`withprime := true;
    new_a := new_a div p;
  end if;

  if not IsSquare(FiniteField(p) ! new_a) then
    rc`nonsquare := true;
  end if;

  return rc;
end intrinsic;


// **************************************************************
// ** Print:
// **************************************************************
intrinsic Print(rc::Rec)
{ print the residue class }
  p := rc`prime;
  if (p eq 0) then
    print "prime:",p,"; nonsquare:",rc`nonsquare;
  elif (p eq 2) then
    print "prime:",p,"; 2:",rc`gener_2,"; -1:",rc`gener_minus1,"; 5:",rc`gener_5;
  else
    print "prime:",p,"; nonsquare:",rc`nonsquare,"; withprime:",rc`withprime;
  end if;
end intrinsic;


// **************************************************************
// ** NormResidueSymbol:
// **************************************************************
intrinsic NormResidueSymbol(
  a::RngIntElt, b::RngIntElt, p::RngIntElt) -> RngIntElt
{ Hilbert Norm Residue Symbol with the given prime }
  a_rc := ResidueClass(a, p);
  b_rc := ResidueClass(b, p);

  return NormResidueSymbol(a_rc, b_rc);
end intrinsic;


// **************************************************************
// ** TestNRS:
// **   Cassels, p.46, Lemma 3.4:
// **   result should be always 1
// **************************************************************
intrinsic TestNRS(a::RngIntElt, b::RngIntElt) -> RngIntElt
{ test routine for Norm residue symbol }
  abs_a := AbsoluteValue(a);
  abs_b := AbsoluteValue(b);
  m := (abs_a lt abs_b) select abs_b else abs_a;

  product := NormResidueSymbol(a,b, 0);

  for i:=1 to m do
    if IsPrime(i) then
      product := product * NormResidueSymbol(a,b, i);
    end if;
  end for;

  return product;
end intrinsic;


// **************************************************************
// ** TestNRSLoop:
// **************************************************************
intrinsic TestNRSLoop(max::RngIntElt)
{ test routine for Norm residue symbol }
  a := [IntegerRing()|];

  repeat
    for i:=1 to 10 do
      a := Random(1, max);
      b := Random(1, max);
      if (Random(1,2) eq 1) then a := -a; end if;
      if (Random(1,2) eq 1) then b := -b; end if;

      product := TestNRS(a,b);
      print product, "for NormResidueSymbol(",a,",",b,")";
    end for;
    print " ";

    read hocico;
  until (hocico eq "q");
end intrinsic;


